/*
 * libvirt-domain-backup.c: entry points for pull backup APIs

 * Copyright (C) 2017 Parallels International GmbH
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include "datatypes.h"
#include "virlog.h"

VIR_LOG_INIT("libvirt.domain-block-snapshot");

#define VIR_FROM_THIS VIR_FROM_DOMAIN

/**
 * virDomainBlockSnapshotXRef:
 * @snapshot: the snapshot to hold a reference on
 *
 * Increment the reference count on the snapshot. For each
 * additional call to this method, there shall be a corresponding
 * call to virDomainBlockSnapshotXFree to release the reference count, once
 * the caller no longer needs the reference to this object.
 *
 * This method is typically useful for applications where multiple
 * threads are using a connection, and it is required that the
 * connection and domain remain open until all threads have finished
 * using the snapshot. ie, each new thread using a snapshot would
 * increment the reference count.
 *
 * Returns 0 in case of success and -1 in case of failure.
 */
int
virDomainBlockSnapshotXRef(virDomainBlockSnapshotXPtr snapshot)
{
    VIR_DEBUG("snapshot=%p, refs=%d", snapshot,
              snapshot ? snapshot->parent.u.s.refs : 0);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, -1);

    virObjectRef(snapshot);
    return 0;
}


/**
 * virDomainBlockSnapshotXFree:
 * @snapshot: a domain snapshot object
 *
 * Free the domain snapshot object. The snapshot itself is not modified.
 * The data structure is freed and should not be used thereafter.
 *
 * Returns 0 in case of success and -1 in case of failure.
 */
int
virDomainBlockSnapshotXFree(virDomainBlockSnapshotXPtr snapshot)
{
    VIR_DEBUG("snapshot=%p", snapshot);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, -1);

    virObjectUnref(snapshot);
    return 0;
}

/**
 * virDomainBlockSnapshotXGetName:
 * @snapshot: a snapshot object
 *
 * Get the name of the snapshot
 *
 * Returns a pointer to the name or NULL, the string need not be deallocated
 * as its lifetime will be the same as the snapshot object.
 */
const char *
virDomainBlockSnapshotXGetName(virDomainBlockSnapshotXPtr snapshot)
{
    VIR_DEBUG("snapshot=%p", snapshot);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, NULL);

    return snapshot->name;
}


/**
 * virDomainBlockSnapshotXGetDomain:
 * @snapshot: a snapshot object
 *
 * Provides the domain pointer associated with a snapshot. The
 * reference counter on the domain is not increased by this
 * call.
 *
 * WARNING: When writing libvirt bindings in other languages, do not use this
 * function. Instead, store the domain and the snapshot object together.
 *
 * Returns the domain or NULL.
 */
virDomainPtr
virDomainBlockSnapshotXGetDomain(virDomainBlockSnapshotXPtr snapshot)
{
    VIR_DEBUG("snapshot=%p", snapshot);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, NULL);

    return snapshot->domain;
}


/**
 * virDomainBlockSnapshotXGetConnect:
 * @snapshot: a snapshot object
 *
 * Provides the connection pointer associated with a snapshot.  The
 * reference counter on the connection is not increased by this
 * call.
 *
 * WARNING: When writing libvirt bindings in other languages, do not use this
 * function. Instead, store the connection and the snapshot object together.
 *
 * Returns the connection or NULL.
 */
virConnectPtr
virDomainBlockSnapshotXGetConnect(virDomainBlockSnapshotXPtr snapshot)
{
    VIR_DEBUG("snapshot=%p", snapshot);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, NULL);

    return snapshot->domain->conn;
}


/**
 * virDomainBlockSnapshotXCreateXML:
 * @domain: a domain object
 * @xmlDesc: block snapshot XML description
 * @flags: bitwise-OR of virDomainBlockSnapshotXCreateFlags
 *
 * Creates block snapshot based on the xml description in
 * @xmlDesc. Block snapshot is different from plain domain disks snapshot.
 * Plain snapshot creates new image file backed by original image file so that all
 * new data being written to disk goes to new image and original image holds
 * data at the moment of the snapshot. Block snaphot creates new image file
 * backed by original image file too but new data still goes to the original
 * image. Before new data get written do the original image the previous data
 * is written to the new image (sometimes called fleece image). As a result
 * fleece image mantain disk snapshot.
 *
 * if @flags includes VIR_DOMAIN_BLOCK_SNAPSHOT_X_CREATE_CHECKPOINT then
 * a checkpoint will be added to every disk in snapshot. Checkpoint provides
 * means to query which disk places are being changed after checkpoint is
 * created. This is typically used for creating incremental backups. For
 * example if first snapshot is created with checkpoint and was copied to
 * backup store then when second snapshot is created backup agent can copy
 * only places that changed from the moment of the first backup.
 *
 * if @flags includes VIR_DOMAIN_BLOCK_SNAPSHOT_X_CREATE_AUTODELETE then
 * block snapshot will be deleted on client disconnect.
 *
 * If @flags includes VIR_DOMAIN_SNAPSHOT_CREATE_QUIESCE, then the
 * libvirt will attempt to use guest agent to freeze and thaw all
 * file systems in use within domain OS. However, if the guest agent
 * is not present, an error is thrown.
 *
 * Returns an (opaque) virDomainBlockSnapshotXPtr on success, NULL on failure.
 */
virDomainBlockSnapshotXPtr
virDomainBlockSnapshotXCreateXML(virDomainPtr domain,
                                 const char *xmlDesc,
                                 unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "xmlDesc=%s, flags=0x%x", xmlDesc, flags);

    virResetLastError();

    virCheckDomainReturn(domain, NULL);
    conn = domain->conn;

    virCheckNonNullArgGoto(xmlDesc, error);
    virCheckReadOnlyGoto(conn->flags, error);

    if (conn->driver->domainBlockSnapshotXCreateXML) {
        virDomainBlockSnapshotXPtr ret;
        ret = conn->driver->domainBlockSnapshotXCreateXML(domain, xmlDesc, flags);
        if (!ret)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return NULL;
}

/**
 * virDomainBlockSnapshotXDelete:
 * @snapshot: a block snapshot object
 * @flags: reserved, must be 0
 *
 * Deletes the snapshot. Fleece images are discared but checkpoints
 * are kept.
 *
 * Returns 0 if the snapshot was successfully deleted, -1 on error.
 */
int
virDomainBlockSnapshotXDelete(virDomainBlockSnapshotXPtr snapshot,
                              unsigned int flags)
{
    virConnectPtr conn;

    VIR_DEBUG("snapshot=%p flags=0x%x", snapshot, flags);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, -1);
    conn = snapshot->domain->conn;

    virCheckReadOnlyGoto(conn->flags, error);

    if (conn->driver->domainBlockSnapshotXDelete) {
        int ret;
        ret = conn->driver->domainBlockSnapshotXDelete(snapshot, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}

/**
 * virDomainBlockSnapshotXList:
 * @domain: a domain object
 * @snaps: pointer to variable to store the array containing block snapshot objects,
 *         or NULL if the list is not required (just returns number of block snapshots)
 * @flags: reserved must be 0
 *
 * Collects the list of block snapshots for the domain, and allocates
 * an array to store those objects.
 *
 * Returns the number of block snapshots found or -1 and sets @snaps to
 * NULL in case of error. On success, the array stored into @snaps is
 * guaranteed to have an extra allocated element set to NULL but not included
 * in the return count, to make iteration easier. The caller is responsible
 * for calling virDomainBlockSnapshotXFree() on each array element, then calling
 * free() on @snaps.
 */
int
virDomainBlockSnapshotXList(virDomainPtr domain,
                            virDomainBlockSnapshotXPtr **snaps,
                            unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "snaps=%p flags=0x%x", snaps, flags);

    virResetLastError();

    virCheckDomainReturn(domain, -1);
    conn = domain->conn;

    if (snaps)
        *snaps = NULL;

    if (conn->driver->domainBlockSnapshotXList) {
        int ret;
        ret = conn->driver->domainBlockSnapshotXList(domain, snaps, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}


/**
 * virDomainBlockSnapshotXGetXMLDesc:
 * @snapshot: a domain snapshot object
 * @flags: reserved, must be 0
 *
 * Provide an XML description of the block snapshot.
 *
 * Returns a 0 terminated UTF-8 encoded XML instance, or NULL in case of error.
 *         the caller must free() the returned value.
 */
char *
virDomainBlockSnapshotXGetXMLDesc(virDomainBlockSnapshotXPtr snapshot,
                                  unsigned int flags)
{
    virConnectPtr conn;

    VIR_DEBUG("snapshot=%p flags=0x%x", snapshot, flags);

    virResetLastError();

    virCheckDomainBlockSnapshotXReturn(snapshot, NULL);
    conn = snapshot->domain->conn;

    if (conn->driver->domainBlockSnapshotXGetXMLDesc) {
        char *ret;
        ret = conn->driver->domainBlockSnapshotXGetXMLDesc(snapshot, flags);
        if (!ret)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return NULL;
}


/**
 * virDomainBlockSnapshotXLookupByName:
 * @domain: a domain object
 * @name: name for the block snapshot
 * @flags: reserved, must be 0
 *
 * Try to lookup a block snapshot based on its name.
 *
 * Returns a block snapshot object or NULL in case of failure. If the
 * domain snapshot cannot be found, then the VIR_ERR_NO_DOMAIN_SNAPSHOT
 * error is raised.
 */
virDomainBlockSnapshotXPtr
virDomainBlockSnapshotXLookupByName(virDomainPtr domain,
                                    const char *name,
                                    unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "name=%s flags=0x%x", name, flags);

    virResetLastError();

    virCheckDomainReturn(domain, NULL);
    conn = domain->conn;

    if (conn->driver->domainBlockSnapshotXLookupByName) {
        virDomainBlockSnapshotXPtr ret;
        ret = conn->driver->domainBlockSnapshotXLookupByName(domain, name, flags);
        if (!ret)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return NULL;
}


/**
 * virDomainBlockCheckpointXRemove:
 * @domain: a domain object
 * @name: a checkpoint name
 * @flags: reserved, must be 0
 *
 * Deletes the checkpoint from all domain disks.
 *
 * Returns 0 if the checkpoint was successfully deleted, -1 on error.
 */
int
virDomainBlockCheckpointXRemove(virDomainPtr domain,
                                const char *name,
                                unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "name=%s flags=0x%x", name, flags);

    virResetLastError();

    virCheckDomainReturn(domain, -1);
    conn = domain->conn;

    virCheckReadOnlyGoto(conn->flags, error);

    if (conn->driver->domainBlockCheckpointXRemove) {
        int ret;
        ret = conn->driver->domainBlockCheckpointXRemove(domain, name, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}


/**
 * virDomainBlockCheckpointXSwitch:
 * @domain: a domain object
 * @name: a checkpoint name
 * @flags: reserved, must be 0
 *
 * Switch to the checkpoint @name. Domain should be in paused start state.
 * Current implementation supports switching to the latest checkpoint only.
 *
 * Returns 0 on success, -1 on error.
 */
int
virDomainBlockCheckpointXSwitch(virDomainPtr domain,
                                const char *name,
                                unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "name=%s flags=0x%x", name, flags);

    virResetLastError();

    virCheckDomainReturn(domain, -1);
    conn = domain->conn;

    virCheckReadOnlyGoto(conn->flags, error);
    virCheckNonNullArgGoto(name, error);

    if (conn->driver->domainBlockCheckpointXSwitch) {
        int ret;
        ret = conn->driver->domainBlockCheckpointXSwitch(domain, name, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}


/**
 * virDomainBlockExportXStart:
 * @domain: a domain object
 * @xmlDesc: export xml description
 * @flags: reserved, must be 0
 *
 * Adds disks/snapshots to export thru specified protocol according to
 * @xmlDesc. Export server will be started if it is not yet running.
 *
 * Returns 0 if the exporst was successfully started, -1 on error.
 */
int
virDomainBlockExportXStart(virDomainPtr domain,
                           const char *xmlDesc,
                           unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "xmlDesc=%s flags=0x%x", xmlDesc, flags);

    virResetLastError();

    virCheckDomainReturn(domain, -1);
    conn = domain->conn;

    virCheckReadOnlyGoto(conn->flags, error);

    if (conn->driver->domainBlockExportXStart) {
        int ret;
        ret = conn->driver->domainBlockExportXStart(domain, xmlDesc, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}


/**
 * virDomainBlockExportXStop:
 * @domain: a domain object
 * @xmlDesc: export xml description
 * @flags: reserved, must be 0
 *
 * Deletes disks/snapshots from export thru specified protocol according to
 * @xmlDesc. Export server will be shutted down if nothing left to export.
 *
 * Returns 0 if the exporst was successfully started, -1 on error.
 */
int
virDomainBlockExportXStop(virDomainPtr domain,
                          const char *xmlDesc,
                          unsigned int flags)
{
    virConnectPtr conn;

    VIR_DOMAIN_DEBUG(domain, "xmlDesc=%s flags=0x%x", xmlDesc, flags);

    virResetLastError();

    virCheckDomainReturn(domain, -1);
    conn = domain->conn;

    virCheckReadOnlyGoto(conn->flags, error);

    if (conn->driver->domainBlockExportXStop) {
        int ret;
        ret = conn->driver->domainBlockExportXStop(domain, xmlDesc, flags);
        if (ret < 0)
            goto error;
        return ret;
    }

    virReportUnsupportedError();
 error:
    virDispatchError(conn);
    return -1;
}
